# -*- coding: cp1250 -*-
# Simulation surface drain
# Created by Bc. Petr Novk, 2015/04
# Faculty of Environmental Sciences
# Czech University of Life Sciences Prague 

"""
Skript obsahuje jednu pomocnou funkci neighbors.
Neighbors hled souadnice sousednch bunk.
Zbvajc st skriptu nate vstupn DEM 
a rastr smr odtok, pevede je na array.
Vytvo jednotkov array, kter me bt voliteln
naven a hodnoty z rastru vah (vlastn, 
nebo zskan metodou Peucker and Douglas, 
viz skript PzD.py).
Sead n-tice hodnot (x,y,z) podle hodnoty z
pomoc prioritn fronty a nsledn prochz 
array pispvajc plochy od msta s nejvy
elevac a navyuje hodnotu elevace bukm 
ve smru odtoku. 
"""

import os, math, arcpy, PaD
import numpy as np
import rta as rt
import Heap as h
arcpy.env.overwriteOutput = True

# Funkce hled [x,y] souadnice sousednch bunk
def neighbors_xy(x,y):
    n1=[x,y+1]
    n2=[x+1,y+1]
    n4=[x+1,y]
    n8=[x+1,y-1]
    n16=[x,y-1]
    n32=[x-1,y-1]
    n64=[x-1,y]
    n128=[x-1,y+1]
    return n1,n2,n4,n8,n16,n32,n64,n128

arcpy.AddMessage("Natm data...")
# Vstupy:
# Rastr DEM
in_dem = sys.argv[1]
# Nate rastr DEM do np.array
aar_dem = rt.rta(in_dem)
ar_dem = aar_dem[0]
XMax = aar_dem[1]
YMax = aar_dem[2]
vCell = aar_dem[3]
LeftX = aar_dem[4]
LextY = aar_dem[5]
del aar_dem

# Rastr smru odtoku
in_dire = sys.argv[2]
# Nate rastr smru odtoku do np.array
ar_dire=rt.rta(in_dire)

# Rastr vhovch koeficient (Peucker and Douglas)
arcpy.AddMessage( sys.argv[5])
if sys.argv[5] == "true":
    # Vypote a nate ve formtu np.array
    ar_scale = PaD.PaD(in_dem)

# Vlastn rastr vhovch koeficient
if sys.argv[6] == "true":
    self_scale = sys.argv[7]
    # Nate rastr vah do np.array
    ar_scale=rt.rta(self_scale)
   
# Vstupy:
# Rastr simulace povrchovho odtoku
out_aku = sys.argv[3]

arcpy.AddMessage( "Data natena, zan vpoet")
ha=[] 
# Vytvo haldu n-tic(x,y,z) azenou podle hodnoty "z"
for xi in range (1,XMax-1):
    for yi in range (1,YMax-1):
        h.inz_el (ha, (xi, yi, ar_dem[xi,yi]), index=2)

# np.array pro hodnoty akumulace
if sys.argv[4]== "true":
    ar_aku=np.ones([XMax,YMax])
else:
    ar_aku=np.ones([XMax,YMax])+ar_scale[0]

arcpy.AddMessage( "Zan vpoet akumulace")
# Hodnota akumulace buky ve smru odtoku
# se nav o hodnotu akumulace ptoku
while len(ha)>0:
    x= ha[0][0]
    y= ha[0][1]
    if x in range (1,XMax-1):
        if y in range (1,YMax-1):             
            sm=ar_dire[0][x][y]
            smer=int(math.log(sm,2))
            nxy=neighbors_xy(x,y)[smer]
            # Buka ve smru odtoku=jej akumulace + akum. ptoku
            ar_aku[nxy[0],nxy[1]]=ar_aku[nxy[0],nxy[1]]+ar_aku[x,y]
    # Odstran max. z haldy a pesune tam dal max.
    h.ret_max(ha,2)

arcpy.AddMessage( "Vypoteno, zapisuji vsledek")
# Vytvo rastr simulace povrchovho odtoku
new_raster = arcpy.NumPyArrayToRaster(ar_aku,arcpy.Point(LeftX,LextY),vCell,value_to_nodata=-9999) 
new_raster.save(out_aku)
