# -*- coding: cp1250 -*-
# River network extraction
# Created by Bc. Petr Novk, 2015/04
# Faculty of Environmental Sciences
# Czech University of Life Sciences Prague 

'''
Skript obsahuje dv pomocn funkce: neighbors_xy a inflov. 
Neighbors_xy hled souadnice sousednch bunk
Inflov zkoum zda je nkter ze sousednch bunk ptokem stedov buky.
Zbvajc st skriptu nate vstupn DEM, rastr smr odtok 
a rastr pispvajcch ploch a pevede je na array. 
Podle zkladn prahov hodnoty pispvajc plochy vyhled msta vzniku 
n st (tzv. prameny), pot vyhled n toky zanajc od pramen.
Nsledn pro jednotliv n dy vypote hodnoty potebn k posouzen 
konstantnho poklesu t-testem, tj. poet tok, prmrn pokles elevac 
a smrodatnou odchylku poklesu elevac. Z tchto hodnot spot T, 
kter mus bt pro vechny po sob jdouc ady tok men ne 1,96. 
Je-li vt, nav se prahov hodnota vzniku n st 
a hledn tok se zopakuje. 
Ve chvli, kdy u je T < 1,96 je vsledn n s uloena do rastru.
'''

import os, math, arcpy
import numpy as np
import rta as rt
arcpy.env.overwriteOutput = True

# Funkce hled [x,y] souadnice sousednch bunk
def neighbors_xy(x,y):
    n1=[x,y+1]
    n2=[x+1,y+1]
    n4=[x+1,y]
    n8=[x+1,y-1]
    n16=[x,y-1]
    n32=[x-1,y-1]
    n64=[x-1,y]
    n128=[x-1,y+1]
    return n1,n2,n4,n8,n16,n32,n64,n128

# Funkce hled ptoky do buky
# Vstupy budou array smr odtok a simulace povrchovho odtoku
def inflow(arr1,arr2,ax,ay):
    p1 = [(ax,  ay+1),16, arr1[0][ax]  [ay+1],arr2[0][ax]  [ay+1]]
    p2 = [(ax+1,ay+1),32, arr1[0][ax+1][ay+1],arr2[0][ax+1][ay+1]]
    p4 = [(ax+1,ay),  64, arr1[0][ax+1][ay],  arr2[0][ax+1][ay]]
    p8 = [(ax+1,ay-1),128,arr1[0][ax+1][ay-1],arr2[0][ax+1][ay-1]]
    p16= [(ax,  ay-1),1,  arr1[0][ax]  [ay-1],arr2[0][ax]  [ay-1]]
    p32= [(ax-1,ay-1),2,  arr1[0][ax-1][ay-1],arr2[0][ax-1][ay-1]]
    p64= [(ax-1,ay),  4,  arr1[0][ax-1][ay],  arr2[0][ax-1][ay]]
    p128=[(ax-1,ay+1),8,  arr1[0][ax-1][ay+1],arr2[0][ax-1][ay+1]]
    # Mon ptoky ze vech smr
    prit = [p1,p2,p4,p8,p16,p32,p64,p128]
    pr=[]
    # Kontrola zda se jedn opravdu o ptok
    for p in prit:
        if p[1]==p[2]:
            if p[3]>=in_ph:
                pr.append(p[0])
    return pr     

arcpy.AddMessage( "Natm data...")
# Vstupy:
# Rastr DEM
in_dem = sys.argv[1]
# Nate rastr DEM do np.array
ar_dem = rt.rta(in_dem)
# ar_dem = aar_dem[0]
XMax = ar_dem[1]
YMax = ar_dem[2]
vCell = ar_dem[3]
LeftX = ar_dem[4]
LextY = ar_dem[5]

# Rastr smr odtok
in_dire = sys.argv[2]
# Nate rastr smr odtok do np.array
ar_dire=rt.rta(in_dire)

# Rastr simulace povrchovho odtoku (pispvajc plochy)
in_aku = sys.argv[3]
# Nate rastr akumulace odtok do np.array
ar_aku=rt.rta(in_aku)
 
# Zkladn prahov hodnota pro extrakci n st
in_ph = int(sys.argv[6])

# Maximln poet cykl hodnocen konstantnho poklesu
ph=[in_ph]
    
# Kroky po kterch se bude zvyovat prahov hodnota
in_phk = XMax*YMax/10000

# Vstupy:
# Rastr n st
out_rs = sys.argv[4]

arcpy.AddMessage("Data natena, zan vpoet")
T=[3]
while max(T) >= 1.96:
    # Hledn pramen (buky s hodnotou akumulace odtoku = prahov hodnot)
    arcpy.AddMessage("Hladn pramen")
    prameny=[] #seznam vech pramen
    # Prameny jsou takov buky, jeji hodnota akumulace je rovna prahov 
    # hodnot, nebo jejich hodnota je vt ne prahov h. a zrove
    # dn z jejich ptok nem hodnotu akumulace vy ne prah.h.
    for x in range (0,XMax-1):
        for y in range (0,YMax-1):
            if ar_aku[0][x][y]==in_ph:
                # Do seznamu pramen pidna (souadnice x,y, Strahlerv ad)
                prameny.append ((x,y,1))
            elif ar_aku[0][x][y] > in_ph and len(inflow(ar_dire, ar_aku,x,y)) == 0:
                # Do seznamu pramen pidna (souadnice x,y, Strahlerv ad)
                prameny.append ((x,y,1))
    arcpy.AddMessage("Nalezeno: "+str(len(prameny))+" pramen")

    #/// Hledn tok (od pramene k soutoku, od soutoku k dalmu soutoku,...)
    arcpy.AddMessage("Hledn tok")
    toky=[] # Seznam vech tok [[I.radu],[II.radu],[III.radu],...]
    dire=rt.rta(in_dire)
    soutoky = [[(XMax,YMax),[]]] # Seznam vech soutok
    #[[(x,y)soutoku,{(x,y,rad)hl.ptoku,(x,y,rad)vedl.pr.,(x,y,rad)..}],[...]]
    while len (prameny) > 0:
        pritok = 0
        x=prameny[0][0]
        y=prameny[0][1]
        rad=prameny[0][2]
        tok=[]
        # Prvn hodnota v seznamu bunk tok je pramen(x,y,rad,z1,zn)
        # [x,y souadnice pramene, . Strahlerova adu, elevace pramene, el. soutoku]
        tok.append ([x,y,rad,ar_dem[0][x][y],ar_dem[0][x][y]])
        while pritok == 0:
            sm=dire[0][x][y]
            # Oeten NoData (-9999.0) ve vrstv smr odtok
            if sm > 0:
                if x in range(1,XMax-2):
                    if y in range(1,YMax-2):
                        # Smr odtoku
                        smer=int(math.log(sm,2))
                        # Souadnice sousedn buky ve smru odtoku
                        nxy=neighbors_xy(x,y)[smer]
                        # Souadnice bunk ptok
                        pxy=inflow(ar_dire, ar_aku, nxy[0],nxy[1])
                        # Je-li v mnoin ptok pouze jedna buka m pouze hl.ptok                   
                        if len (pxy)==1:
                            # Posun se o buku dl
                            x=nxy[0]
                            y=nxy[1]
                            # Pidm souadnice dal buky toku do seznamu
                            tok.append (nxy)
                            tok[0][4]=ar_dem[0][x][y]
                        # Obsahuje-li v mnoina ptok vce buek jedn se o soutok
                        elif len (pxy)> 1:
                            # Pid souadnice soutoku do seznamu bunk toku
                            tok[0][4] = ar_dem[0][nxy[0]][nxy[1]]
                            tok.append (nxy)
                            # Prohled vechny ptoky
                            for p in pxy:
                                if p==(x,y): # hl. ptok
                                    pass
                                else: # Vedlej ptok
                                    # Prohled seznam soutok, jestli tam tento soutok u nen
                                    sindex= 0
                                    for i in soutoky:
                                        if (nxy[0],nxy[1]) in i:
                                            sindex=soutoky.index (i)
                                    # Soutok je ji v seznamu soutok                                      
                                    if (nxy[0],nxy[1]) in soutoky[int(sindex)]: 
                                        # Pid mu aktuln hl. ptok
                                        soutoky[sindex][1].append ((x,y,rad)) 
                                        so=set()
                                        for s in soutoky[sindex][1]:
                                            so.add ((s[0],s[1]))
                                        # Je-li seznam ptok v seznamu soutok shodn s ptoky buky    
                                        if so==set(pxy): 
                                            # Zjist ad toku pod soutokem
                                            rd=[] # Seznam ad vech ptok
                                            for ra in soutoky[sindex][1]:
                                                rd.append (ra[2])
                                            # Najde dva nejvy ady    
                                            rd.sort(reverse=True)
                                            # Jsou-li stejn nsledujc je o jedno vy 
                                            if rd[0]==rd[1]: 
                                                srad=rd[0]+1
                                            # Jsou-li rozdln nsledujc m hodnotu vyho z ptok
                                            elif rd[0]>rd[1]: 
                                                srad=rd[0]
                                            else:
                                                srad=rd[1]
                                            # Pid soutok do seznamu pramen
                                            prameny.append ((nxy[0],nxy[1],srad))
                                            # a vymae tento soutok ze seznamu soutoku 
                                            del soutoky[sindex]
                                        # Nejsou-li mnoiny ptok shodn pokrauje znovu od dalho pramene 
                                        else: 
                                            pass
                                    # Pid souadnic soutoku do seznamu soutoku         
                                    else:                                    
                                        soutoky.append ([(nxy[0],nxy[1]),[(x,y,rad)]])
                                        pritok=1
                                    pritok=1
                        else:
                            arcpy.AddError("Chyba v akumulaci toku!!!")
                            akum_oprava=[]
                            pritok=1
                    else:
                        pritok=1                    
                else:
                    pritok=1               
            else:
                pritok=1                
        # Analyzovan tok je pidn do seznamu tok
        if len(toky) < rad:
            toky.append ([tok])
        else:
            toky[rad-1].append (tok)
        del prameny[0] 

    # Hodnocen konstantnho poklesu
    if sys.argv[5] == 'true':
        # Posouzeni "konstantnho poklesu" 
        # Vyhov-li vytvo se novy rastr,
        # nevyhov-li ur se nova prahov hodnota a vpoet se zopakuje.
        # Tok:[x,y,rad,z1,zn]
        arcpy.AddMessage("Hodnoceni konstantnho poklesu")
        #arcpy.AddMessage(str(toky))
        # Vpoet prmrnho rozdlu vek a potu prvk
        hodnoty=[]#[[i,ni,pzi,rsi],[]]
        for i in range (len(toky)):
            ni=len(toky[i])
            zi=0
            pzi=0
            for j in range (len(toky[i])):
                zi=zi+((toky[i][j][0][3]-toky[i][j][0][4]))
            pzi=zi/ni
            hodnoty.append([i+1, ni, pzi])
            
        # Vpoet smrodatn odchylky    
        for i in range(len(toky)):
            si=0
            for j in range(len(toky[i])):        
                ni=hodnoty[i-1][1]
                pzi= hodnoty[i-1][2]
                si=si+(((toky[i][j][0][3]-toky[i][j][0][4])-hodnoty[i-1][2])**2)
            rsi=si/(ni-1)
            hodnoty[i-1].append(rsi)
        #arcpy.AddMessage(hodnoty)   
        # Vpoet t-testu
        T=[]
        for i in range(len(hodnoty)-1):
            Tt=(((hodnoty[i][2])-(hodnoty[i+1][2]))/(math.sqrt((((hodnoty[i][1]-1)*((hodnoty[i][3])**2))+((hodnoty[i+1][1]-1)*((hodnoty[i+1][3])**2)))/((hodnoty[i][1])+(hodnoty[i+1][1])-2))*math.sqrt((1.0/(hodnoty[i][1]))+(1.0/(hodnoty[i+1][1])))))
            T.append(Tt)
        #arcpy.AddMessage("T: "+str(T))
        if max(T)>=1.96:
            if len (ph)>50:
                arcpy.AddError("Vypocet byl po 50 cyklech ukonen, nebyl nalezen korektn vsledek!!!")
                T=[0]
            in_ph=in_ph+in_phk
            ph.append (in_ph)
            arcpy.AddMessage("Prahov hodnota navena na "+str(in_ph))          
    else:
        T=[0]
        arcpy.AddMessage("Vpoet n st pro prahovou hodnotu "+str(in_ph)+"je hotov.") 

arcpy.AddMessage("Zapisuji vsledek")
# np.array pro hodnoty vsledn vrstvy n st
ar_rs=np.zeros([XMax,YMax])
for r in toky:
    for t in r:
        for body in t:
            ar_rs[body[0],body[1]] = t[0][2]
# Vytvo rastr n st 
new_raster = arcpy.NumPyArrayToRaster(ar_rs,arcpy.Point(LeftX,LextY),vCell,value_to_nodata=-9999) 
new_raster.save(out_rs)


